      SUBROUTINE gaussj(a,n,np,b,m,mp)
      INTEGER m,mp,n,np,NMAX
      REAL*8 a(np,np),b(np,mp)
      PARAMETER (NMAX=50)
      INTEGER i,icol,irow,j,k,l,ll,indxc(NMAX),indxr(NMAX),ipiv(NMAX) 
C     The integer arrays ipiv, indxr, and indxc are used
C     for bookkeeping on the pivoting.
      REAL*8 big,dum,pivinv

      do j=1,n
         ipiv(j)=0
      enddo 
      do i=1,n !This is the main loop over the columns to be reduced.
         big=0.d0
         do j=1,n !This is the outer loop of the search for a pivot element.
            if(ipiv(j).ne.1)then
               do k=1,n
                  if (ipiv(k).eq.0) then
                     if (abs(a(j,k)).ge.big)then
                        big=abs(a(j,k))
                        irow=j
                        icol=k
                     endif
                  else if (ipiv(k).gt.1) then
                     pause 'singular matrix in gaussj'
                  endif
               enddo 
            endif
         enddo 
         ipiv(icol)=ipiv(icol)+1
C     We now have the pivot element, so we interchange rows, if needed, 
C     to put the pivot element on the diagonal. The columns are not 
C     physically interchanged, only relabeled:

C     indxc(i), the column of the ith pivot element, is the ith column 
C     that is reduced, while
C     indxr(i) is the row in which that pivot element was originally located. 
C     If indxr(i)6.ne.indxc(i) there is an implied column interchange. 
C     With this form of bookkeeping, the solution b's will end up in the 
C     correct order, and the inverse matrix will be scrambled by columns.
         if (irow.ne.icol) then
            do l=1,n
               dum=a(irow,l)
               a(irow,l)=a(icol,l)
               a(icol,l)=dum
            enddo 
            do l=1,m
               dum=b(irow,l)
               b(irow,l)=b(icol,l)
               b(icol,l)=dum
            enddo 
         endif
         indxr(i)=irow !We are now ready to divide the pivot row by the pivot element, located at irow and icol.
         indxc(i)=icol
         if (a(icol,icol).eq.0.) pause 'singular matrix in gaussj'
         pivinv=1.d0/a(icol,icol)
         a(icol,icol)=1.d0
         do l=1,n
            a(icol,l)=a(icol,l)*pivinv
         enddo 
         do l=1,m
            b(icol,l)=b(icol,l)*pivinv
         enddo 
         do ll=1,n !Next, we reduce the rows...
            if(ll.ne.icol)then !...except for the pivot one, of course.
               dum=a(ll,icol)
               a(ll,icol)=0.d0
               do l=1,n
                  a(ll,l)=a(ll,l)-a(icol,l)*dum
               enddo 
               do l=1,m
                  b(ll,l)=b(ll,l)-b(icol,l)*dum
               enddo 
            endif
         enddo 
      enddo !This is the end of the main loop over columns of the reduction.
      do l=n,1,-1  
C     It only remains to unscramble the solution in view
C     of the column interchanges. We do this by in-
C     terchanging pairs of columns in the reverse order
C     that the permutation was built up.
         if(indxr(l).ne.indxc(l))then
            do k=1,n
               dum=a(k,indxr(l))
               a(k,indxr(l))=a(k,indxc(l))
               a(k,indxc(l))=dum
            enddo 
         endif
      enddo 
      return !And we are done.
      END
